/* osd.c - OSD implementation using XOSD
 *
 * Copyright 2009, 2015, 2022, 2023 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of xosdd.
 *
 * xosdd is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * xosdd is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with xosdd.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "osd.h"
#include "util.h"
#include <xosd.h>
#include <assert.h>
#include <errno.h>
#include <string.h>

struct osd {
    xosd* osd;
};

static xosd_pos osd_position;

static int str_to_align(const char *s, xosd_align *align);
static int str_to_pos(const char *s, xosd_pos *pos);

osd*
create_osd(int nlines)
{
    osd *osd = ph_malloc(sizeof(osd));
    osd->osd = xosd_create(nlines);
    if (!osd->osd) {
        ph_eprintf("can't create OSD: %s", xosd_error);
        return NULL;
    } else {
        return osd;
    }
}

void
destroy_osd(osd *osd)
{
    int rval = xosd_destroy(osd->osd);
    if (rval != 0) ph_eprintf("can't destroy OSD: %s", xosd_error);
    free(osd);
}

/* Hide OSD. */
void
hide_osd(osd *osd)
{
    int rval = xosd_hide(osd->osd);
    if (rval == -1) ph_eprintf("can't hide OSD");
}

/* Display percentage val on OSD at the given line with indicator ind.
 * Clamp line between 0 and the number of the lines in OSD minus one,
 * and clamp val between 0 and 100. */
void
osd_display(osd *osd, int line, int val, enum indicator ind)
{
    int nlines;
    int rval;

    nlines = xosd_get_number_lines(osd->osd);
    if (nlines == -1) {
        ph_wprintf("can't get the number of lines in the OSD.");
        return;
    }
    line = ph_clamp_int(line, 0, nlines - 1);

    val = ph_clamp_int(val, 0, 100);
    switch (ind) {
    case BAR:
        rval = xosd_display(osd->osd, line, XOSD_percentage, val);
        break;
    case SLIDER:
        rval = xosd_display(osd->osd, line, XOSD_slider, val);
        break;
    default:
        rval = -1;
    }
    if (rval == -1) {
        ph_eprintf("can't display value %d on OSD line %d: %s",
                   val, line, xosd_error);
    }
}

/* Display s on OSD at the given line.  Clamp line between 0 and the
 * number of the lines in OSD minus one. */
void
osd_string(osd *osd, int line, const char *s)
{
    int nlines;
    int rval;

    nlines = xosd_get_number_lines(osd->osd);
    if (nlines == -1) {
        ph_wprintf("can't get the number of lines in the OSD.");
        return;
    }
    line = ph_clamp_int(line, 0, nlines - 1);

    rval = xosd_display(osd->osd, line, XOSD_string, s);
    if (rval == -1) {
        ph_eprintf("can't display string '%s' on OSD line %d: %s",
                   s, line, xosd_error);
    }
}

/* Display strings, one string per line.  strings array must be
 * terminated by a null pointer. */
void
osd_strings(osd *osd, char **strings)
{
    int rval;
    int line;
    int i;
    int step;
    int nlines;

    nlines = xosd_get_number_lines(osd->osd);
    if (nlines == -1) {
        ph_eprintf("can't get the number of lines in the OSD.");
        return;
    }

    if (osd_position == XOSD_top) {
        line = 0;
        step = 1;
    } else {
        line = nlines - 1;
        step = -1;
    }

    /* display strings */
    i = 0;
    while (strings[i] && line >= 0 && line < nlines) {
        rval = xosd_display(osd->osd, line, XOSD_string, strings[i]);
        if (rval == -1) {
            ph_eprintf("can't display string '%s' on OSD line %d: %s",
                       strings[i], line, xosd_error);
        }
        line += step;
        i++;
    }

    /* clear remaining lines */
    while (line >= 0 && line < nlines) {
        ph_dprintf("clearing %d", line);
        rval = xosd_display(osd->osd, line, XOSD_string, '\0');
        if (rval == -1) {
            ph_eprintf("can't clear OSD line %d: %s",
                       line, xosd_error);
        }
        line += step;
    }
}

/* Display percentage val (with indicator ind) and strings.  strings
 * array must be terminated by a null pointer.
 *
 * If the OSD is at the top of the screen, the value indicator is
 * displayed at the top, and the strings appear below it in ascending
 * order.  Otherwise, if the OSD is at the bottom of the screen, we do
 * exactly the opposite. */
void
osd_value_and_strings(osd *osd, int val, char **strings,
                      enum indicator ind)
{
    int rval;
    int line;
    int i;
    int step;
    int nlines;

    nlines = xosd_get_number_lines(osd->osd);
    if (nlines == -1) {
        ph_eprintf("can't get the number of lines in the OSD.");
        return;
    }

    if (osd_position == XOSD_top) {
        line = 0;
        step = 1;
    } else {
        line = nlines - 1;
        step = -1;
    }

    /* display value */
    val = ph_clamp_int(val, 0, 100);
    switch (ind) {
    case BAR:
        rval = xosd_display(osd->osd, line, XOSD_percentage, val);
        break;
    case SLIDER:
        rval = xosd_display(osd->osd, line, XOSD_slider, val);
        break;
    default:
        rval = -1;
    }
    if (rval == -1) {
        ph_eprintf("can't display value %d: %s", val, xosd_error);
    }

    /* display strings */
    line += step;
    i = 0;
    while (strings[i] && line >= 0 && line < nlines) {
        rval = xosd_display(osd->osd, line, XOSD_string, strings[i]);
        if (rval == -1) {
            ph_eprintf("can't display string '%s' on OSD line %d: %s",
                       strings[i], line, xosd_error);
        }
        i++;
        line += step;
    }

    /* clear remaining lines */
    while (line >= 0 && line < nlines) {
        rval = xosd_display(osd->osd, line, XOSD_string, '\0');
        if (rval == -1) {
            ph_eprintf("can't clear OSD line %d: %s",
                       line, xosd_error);
        }
        line += step;
    }
}

/* Scroll OSD by the given number of lines.  Clamp lines between 0 and
 * the number of the lines in OSD minus one. */
void
scroll_osd(osd *osd, int lines)
{
    int rval;
    int nlines;

    nlines = xosd_get_number_lines(osd->osd);
    if (nlines == -1) {
        ph_eprintf("can't get the number of lines in the OSD.");
        return;
    }

    lines = ph_clamp_int(lines, 0, nlines - 1);
    rval = xosd_scroll(osd->osd, lines);
    if (rval == -1) {
        ph_eprintf("can't scroll %d lines", lines);
    }
}

/* Set OSD horizontal alignment. */
void
set_align(osd *osd, const char *align)
{
    xosd_align x_align;
    if (str_to_align(align, &x_align) == EINVAL) {
        return;
    }
    else {
        int rval = xosd_set_align(osd->osd, x_align);
        if (rval == -1) {
            ph_eprintf("can't set alignment");
        }
    }
}

/* Set OSD bar length. */
void
set_bar_length(osd *osd, int len)
{
    int rval = xosd_set_bar_length(osd->osd, len);
    if (rval == -1) {
        ph_eprintf("can't set bar length to %d: %s",
                   len, xosd_error);
    }
}

/* Set OSD foreground color. */
void
set_color(osd *osd, const char *color)
{
    int rval = xosd_set_colour(osd->osd, color);
    if (rval == -1) {
        ph_eprintf("can't set osd color to '%s'", color);
    }
}

/* Set OSD font. */
void
set_font(osd *osd, const char *font)
{
    int rval = xosd_set_font(osd->osd, font);
    if (rval == -1) {
        ph_eprintf("can't set font to '%s'", font);
    }
}

/* Set OSD horizontal offset (in pixels). */
void
set_horizontal_offset(osd *osd, int offset)
{
    int rval = xosd_set_horizontal_offset(osd->osd, offset);
    if (rval == -1) {
        ph_eprintf("can't set horizontal offset to %d", offset);
    }
}

/* Set OSD outline color. */
void
set_outline_color(osd *osd, const char *color)
{
    int rval = xosd_set_outline_colour(osd->osd, color);
    if (rval == -1) {
        ph_eprintf("can't set outline color to '%s'", color);
    }
}

/* Set OSD outline offset (in pixels). */
void
set_outline_offset(osd *osd, int offset)
{
    int rval = xosd_set_outline_offset(osd->osd, offset);
    if (rval == -1) {
        ph_eprintf("can't set outline offset to %d", offset);
    }
}

/* Set OSD vertical position and update osd_position. */
void
set_position(osd *osd, const char *pos)
{
    xosd_pos x_pos;
    if (str_to_pos(pos, &x_pos) == EINVAL) {
        return;
    } else {
        int rval = xosd_set_pos(osd->osd, x_pos);
        if (rval == -1) {
            ph_eprintf("can't set position");
        } else {
            osd_position = x_pos;
        }
    }
}

/* Set OSD shadow offset (in pixels). */
void
set_shadow_offset(osd *osd, int offset)
{
    int rval = xosd_set_shadow_offset(osd->osd, offset);
    if (rval == -1) {
        ph_eprintf("can't set shadow offset to %d", offset);
    }
}

/* Set OSD timeout (in seconds). */
void
set_timeout(osd *osd, int timeout)
{
    int rval = xosd_set_timeout(osd->osd, timeout);
    if (rval == -1) {
        ph_eprintf("can't set timeout to %d", timeout);
    }
}

/* Set OSD vertical offset (in pixels). */
void
set_vertical_offset(osd *osd, int offset)
{
    int rval = xosd_set_vertical_offset(osd->osd, offset);
    if (rval == -1) {
        ph_eprintf("can't set vertical offset to %d", offset);
    }
}

/* Show OSD. */
void
show_osd(osd *osd)
{
    int rval = xosd_show(osd->osd);
    if (rval == -1) ph_eprintf("can't show OSD");
}

/* Convert s to XOSD alignment pointed by align.  Return 0 on success,
 * EINVAL otherwise. */
static int
str_to_align(const char *s, xosd_align *align)
{
    assert(s && align);
    if (strcmp(s, "left") == 0) {
        *align = XOSD_left;
    } else if (strcmp(s, "center") == 0) {
        *align = XOSD_center;
    } else if (strcmp(s, "right") == 0) {
        *align = XOSD_right;
    } else {
        ph_eprintf("can't convert '%s' to alignment", s);
        return EINVAL;
    }
    return 0;
}

/* Convert s to XOSD position pointed by pos.  Return 0 on success,
 * EINVAL otherwise. */
static int
str_to_pos(const char *s, xosd_pos *pos)
{
    assert(s && pos);
    if (strcmp(s, "top") == 0) {
        *pos = XOSD_top;
    } else if (strcmp(s, "bottom") == 0) {
        *pos = XOSD_bottom;
    } else {
        ph_eprintf("can't convert '%s' to position", s);
        return EINVAL;
    }
    return 0;
}
