/* util.c - miscellaneous utility functions
 *
 * These functions are generally not re-entrant, and they are meant
 * for static linking.  They should be C89-compliant.
 *
 * Copyright 2009, 2015, 2022, 2023 Petteri Hintsanen <petterih@iki.fi>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see
 * <http://www.gnu.org/licenses/>. 
 */

#include "util.h"
#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

enum msg_type { DEBUG, ERROR, FATAL_ERROR, OPTION_ERROR, LOG,
                WARNING };

static char **add_token(char ***tokens, int *ntokens, int *nrestok,
                        const char *token);
static char *append(char **s, char **end, size_t *size,
                    const char c);
static void print(enum msg_type t, unsigned int mask, const char *fmt,
                  va_list ap);

static char *program_name = NULL;
static int debug_level = 0;
static int log_level = 0;

/* Add a copy of token into tokens table, grow it if necessary.
 * ntokens must contain the number of tokens in the table, and nres
 * must contain the number of allocated table slots, including the
 * terminating null token.  Return a (possibly relocated) pointer to
 * the tokens table. */
static char **
add_token(char ***tokens, int *ntokens, int *nres, const char *token)
{
    (*ntokens)++;
    if (*ntokens > *nres) {
        *nres *= 2;
        /* ph_dprintf("reallocating %d bytes for token table",  */
        /*           *nres * sizeof(char *)); */
        *tokens = ph_realloc(*tokens, *nres * sizeof(char *));
    }

    (*tokens)[*ntokens - 2] = ph_strdup(token);
    /* ph_dprintf("found token '%s'", (*tokens)[*ntokens - 2]); */
    return *tokens;
}

/* Append c to s (grow s if necessary), and terminate s with the null
 * byte.  end must point to the terminating null byte of s.  size must
 * be the size of the memory block allocated for s, including the
 * terminating null byte.  Return a (possibly relocated) pointer to
 * s. */
static char *
append(char **s, char **end, size_t *size, const char c)
{
    /* ph_dprintf("appending '%c' to '%s'", c, *s); */

    /* enlarge buffer if needed */
    size_t len = *end - *s + 1;
    /* ph_dprintf("length is %d bytes", tokenlen); */
    if (len + 1 > *size) {
        *size *= 2;
        /* ph_dprintf("reallocating %d bytes for token buffer", *bufsize); */
        *s = ph_realloc(*s, *size);
        *end = *s + len - 1;
    }

    **end = c;
    (*end)++;
    **end = '\0';
    return *s;
}

/* Print formatted message (fmt and ap) to stderr with prefix.  If t
 * is DEBUG, consult debug_level to see whether to print anything at
 * all:
 *
 * - If debug_level is zero, return without printing anything.
 *
 * - If debug_level is negative, print regardless of mask.
 *
 * - Otherwise, print if debug_level & mask is nonzero.
 *
 * If t is LOG, do the same as above, but consult log_level
 * instead. */
static void
print(enum msg_type t, unsigned int mask, const char *fmt, va_list ap)
{
    /* fprintf(stderr, "print: debug_level %d log_level %d mask %d\n", */
    /*         debug_level, log_level, mask); */

    /* check if we ought to print */
    if (t == LOG
        && (log_level == 0
            || (log_level >= 0 && mask > 0
                && (log_level & mask) == 0))) return;
    else if (t == DEBUG
             && (debug_level == 0
                 || (debug_level >= 0 && mask > 0
                     && (debug_level & mask) == 0))) return;

    if (program_name) fprintf(stderr, "%s: ", program_name);

    switch (t) {
    case DEBUG:
        fprintf(stderr, "debug: ");
        break;
    case ERROR:
        fprintf(stderr, "error: ");
        break;
    case FATAL_ERROR:
        fprintf(stderr, "fatal error: ");
        break;
    case WARNING:
        fprintf(stderr, "warning: ");
        break;
    default:
        break;
    }

    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\n");

    if (t == OPTION_ERROR) {
        fprintf(stderr, "Try '%s -h' for more information.\n",
                program_name);
    }
}

/* Return a pointer to the beginning of the filename part in path. */
const char *
ph_basename(const char *path)
{
    char *s;
    assert(path);
    s = strrchr(path, '/');
    if (s == NULL) return path;
    else return s + 1;
}

/* Generate a Bernoulli random variable with parameter p. */
int
ph_bernoulli(double p)
{
    return 1 - (int) (rand() / (RAND_MAX + 1.0) / p);
}

/* Clamp val to the specified range.  If val is less than min, return
 * min.  If val is greater than max, return max.  Otherwise, return
 * val. */
int
ph_clamp_int(int val, int min, int max)
{
    if (val < min) val = min;
    else if (val > max) val = max;
    return val;
}

/* Print formatted debugging message to stderr.  This is a convenience
 * function which prints message with zero mask, so that the message
 * will be printed unless debug_level is zero. */
void
ph_dprintf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    print(DEBUG, 0, fmt, ap);
    va_end(ap);
}

/* Print formatted debugging message to stderr with given mask.  See
 * also ph_dprintf. */
void
ph_dmprintf(unsigned int mask, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    print(DEBUG, mask, fmt, ap);
    va_end(ap);
}

/* Print formatted error message to stderr. */
void
ph_eprintf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    print(ERROR, 0, fmt, ap);
    va_end(ap);
}

/* Print formatted fatal error message to stderr and exit with
 * failure. */
void
ph_feprintf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    print(FATAL_ERROR, 0, fmt, ap);
    va_end(ap);
    exit(EXIT_FAILURE);
}

/* Free memory allocated for the tokens in the given array, as well as
 * the array itself.  See ph_strtok. 
 *
 * tokens can be NULL, in which case no operation is performed. */
void
ph_free_tokens(char **tokens)
{
    char **token;

    if (!tokens) return;
    
    token = tokens;
    while (*token) {
        /* ph_dprintf("  freeing '%s'", *token); */
        free(*token);
        token++;
    }

    /* ph_dprintf("  freeing tokens array"); */
    free(tokens);
}

/* Read characters from stream into a string until a newline character
 * is seen or the end of stream is reached.  The newline character is
 * discarded; it is not a part of the returned string.  Allocate
 * memory for the string holding the line, and return a pointer to
 * it. */
char *
ph_getline(FILE *stream)
{
    int ch;
    int nchars = 0;
    char *buf;
    size_t nres = 256;

    assert(stream);
    if (feof(stream)) return NULL;

    buf = (char*) ph_malloc(nres + 1);
    while ((ch = getc(stream)) && ch != '\n' && ch != EOF) {
        assert(nchars <= nres);
        if (nchars == nres) {
            nres *= 2;
            buf = ph_realloc(buf, nres + 1);
        }
        buf[nchars] = ch;
        nchars++;
    }

    if (nchars == 0 && ch == EOF) {
        free(buf);
        return NULL;
    } else {
        buf[nchars] = '\0';
        return buf;
    }
}

/* Print formatted log message to stderr.  This is a convenience
 * function which prints message with zero mask, so that the message
 * will be printed unless log_level is zero. */
void ph_lprintf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    print(LOG, 0, fmt, ap);
    va_end(ap);
}

/* Print formatted log message to stderr with given mask.
 * See also ph_lprintf. */
void ph_lmprintf(unsigned int mask, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    print(LOG, mask, fmt, ap);
    va_end(ap);
}

/* Try to allocate size bytes of memory with malloc.  If size is zero,
 * return NULL.  If allocation fails, terminate the program with fatal
 * error. */
void *
ph_malloc(size_t size)
{
    void *ptr;
    if (size == 0) return NULL;
    ptr = malloc(size);
    if (!ptr) ph_feprintf("out of memory");
    return ptr;
}

/* Print formatted option error string, advice user to try '-h' switch
 * to see program usage, and exit with failure.  This function is
 * intended to be called from an option parser. */
void
ph_oeprintf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    print(OPTION_ERROR, 0, fmt, ap);
    exit(EXIT_FAILURE);
}

/* Return the current program name.  Program name is used in
 * diagnostic message functions (ph_...printf). */
const char *
ph_program_name(void)
{
    return program_name;
}

/* Try to reallocate ptr to hold size bytes of memory.  If ptr is NULL
 * and size is zero, return null.  If reallocation fails, terminate
 * the program with fatal error. */
void *
ph_realloc(void *ptr, size_t size)
{
    if (ptr == NULL && size == 0) return NULL;
    ptr = realloc(ptr, size);
    if (!ptr) ph_feprintf("out of memory");
    return ptr;
}

/* Set the current debug level.  Debug level is used to control output
 * from ph_dprintf and ph_dlprintf.  In general, if debug_level is
 * zero, no debugging output is printed.  If debug_level is negative,
 * full debugging output will be printed. */
void
ph_set_debug_level(int level)
{
    debug_level = level;
}

/* Set the current log level.  Log level is used to control output
 * from ph_lprintf and ph_llprintf.  In general, if log_level is zero,
 * no logging is produced.  If log_level is negative, full logs will
 * be printed. */
void
ph_set_log_level(int level)
{
    log_level = level;
}

/* Set the current program name. Program name is used in diagnostic
 * message functions (ph_...printf). */
void
ph_set_program_name(const char *name)
{
    if (program_name) free(program_name);
    program_name = ph_strdup(name);
}

/* Initialize random number generator with the given seed.  If seed is
 * zero, use time(3) as seed. */
void
ph_srand(unsigned int seed)
{
    if (seed == 0) srand((int) time(NULL));
    else srand(seed);
}

/* Concatenate a null-terminated array of strings with delimiter
 * delim.  Allocate memory for the concatenated string and return a
 * pointer to it.
 *
 * If delimiter is the null byte, do not use any delimiter.  This is
 * similar to what strcat(3) would do.
 */
char *
ph_strscat(const char **strings, char delim)
{
    int i, len, pad;
    char *res, *sp;

    assert(strings);

    i = 0;
    len = 0;
    pad = 0;
    if (delim != '\0') pad = 1;
    while (strings[i]) {
        /* ph_dprintf("handling '%s'", strings[i]); */
        len += (strlen(strings[i]) + pad);
        i++;
    }
    if (delim != '\0') len++; /* Reserve space for the terminating
                               * null byte. */
    /* ph_dprintf("total string length is %d", len); */
    if (len == 0) return NULL;

    res = ph_malloc(len);
    sp = res;
    i = 0;
    while (strings[i]) {
        /* ph_dprintf("copying '%s' to %p", strings[i], sp); */
        memcpy(sp, strings[i], strlen(strings[i]));
        sp += strlen(strings[i++]);
        if (delim != '\0') *sp++ = delim;
    }
    if (delim != '\0') sp--;
    *sp = '\0';

    /* ph_dprintf("final string '%s'", res); */
    return res;
}

/* Duplicate s.  Allocate memory for the copy and return a pointer to
 * it. */
char *
ph_strdup(const char *s)
{
    char *dest;
    assert(s);
    dest = ph_malloc(strlen(s) + 1);
    strcpy(dest, s);
    return dest;
}

/* Tokenize s by extracting substrings (tokens) separated by any
 * character in seps.  Tokens can be quoted with the given quote
 * character, which can be represented (escaped) by doubling it within
 * a token.
 *
 * Empty tokens are returned, unlike C89 strtok(3).
 *
 * Return an array of tokens, terminated by a null token.  
 * If s in NULL, return NULL.
 *
 * You should use free_tokens to free the memory reserved for the
 * tokens and the array after you have finished with the tokens.
 *
 * Example:
 *
 * string s: "foo ""bar" baz , quux" "bz
 * seps: " ," (space and comma)
 * quote: " (double quote)
 * gives five tokens:
 *   foo "bar
 *   baz
 *   (empty token)
 *   (empty token)
 *   quux bz
 */
char **
ph_strtok(const char *s, const char *seps, const char quote)
{
    const char *c;
    char *token;
    char *end;
    size_t bufsize = 256;

    char **tokens;
    int ntokens = 1;            /* initially only the terminating null
                                 * token */
    int nrestok = 16;

    typedef enum { UNQUOTED, QUOTED, ESCAPE, UNDEFINED } State; /* DFA state */
    State state;                /* current state */
    State last;                 /* state before quote char */

    if (s == NULL) return NULL;

    /* ph_dprintf("tokenizing '%s'", str); */

    token = ph_malloc(bufsize);
    tokens = ph_malloc(nrestok * sizeof(char *));
    c = s;
    state = UNQUOTED;
    last = UNDEFINED;
    end = token;
    *end = '\0';

    while (*c) {
        switch (state) {
        case UNQUOTED:
            if (*c == quote) {
                /* check for escaped quote */
                last = UNQUOTED;
                state = ESCAPE;
            } else if (strchr(seps, *c)) {
                /* end of token, store it and begin new */
                add_token(&tokens, &ntokens, &nrestok, token);
                end = token;
                *end = '\0';
            } else {
                /* regular character */
                append(&token, &end, &bufsize, *c);
            }
            c++;
            break;

        case QUOTED:
            if (*c == quote) {
                /* check for escaped quote */
                last = QUOTED;
                state = ESCAPE;
            } else {
                /* non-quote character */
                append(&token, &end, &bufsize, *c);
            }
            c++;
            break;

        case ESCAPE:
            if (*c == quote) {
                /* escaped quote, add it and return to the last
                 * non-escape state */
                append(&token, &end, &bufsize, *c);
                state = last;
            } else if (strchr(seps, *c) && last == QUOTED) {
                /* end of a quoted token, store it and begin new */
                add_token(&tokens, &ntokens, &nrestok, token);
                end = token;
                *end = '\0';
                state = UNQUOTED;
            } else {
                /* quoted sequence start/end */
                append(&token, &end, &bufsize, *c);
                if (last == UNQUOTED) {
                    state = QUOTED;
                } else {
                    state = UNQUOTED;
                }
            }
            c++;
            break;
        case UNDEFINED:
            assert(0);
        }
    }

    /* store the last token */
    add_token(&tokens, &ntokens, &nrestok, token);

    tokens[ntokens - 1] = '\0';

    /* ph_dprintf("found %d tokens:", ntokens); */
    /* for (i = 0; i < ntokens; i++) { */
    /*     ph_dprintf("  %d: '%s'", i, tokens[i]); */
    /* } */

    free(token);
    return tokens;
}

/* Convert s to an integer pointed by i.  Return 0 on success, EINVAL
 * otherwise. */
int
ph_strtoi(const char *s, int *i)
{
    char *end;

    assert(s && i);

    *i = strtol(s, &end, 10);
    if (end == s || (*i == 0 && errno == EINVAL)) return EINVAL;
    else return 0;
}

/* Print formatted warning message to stderr. */
void
ph_wprintf(const char *fmt, ...)
{
    va_list ap;
    assert(fmt);
    va_start(ap, fmt);
    print(WARNING, 0, fmt, ap);
    va_end(ap);
}
